# How to run OctaDist
# 1. Stay outside octadist directory
# 2. Execute following command:
#    $ python -m octadist.Run

import octadist.main

version = octadist.__version__
release = octadist.__release__

if __name__ == '__main__':
    print(f"\nProgram Started >>>")
    print(f"... OctaDist {version} {release} ...")

    octadist.main.main()

    print(f"<<< Program Terminated")



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np


def find_eq_of_plane(x, y, z):
    """
    Find the equation of plane of given three points using cross product

    | The general form of plane equation:
    |
    | Ax + By + Cz = D
    |
    | where A, B, C, and D are coefficient.
    |
    | XZ  X  XY = (a, b, c)
    |
    | d = (a, b, c).Z

    Parameters
    ----------
    x : array_like
        3D Coordinate of point.
    y : array_like
        3D Coordinate of point.
    z : array_like
        3D Coordinate of point.

    Returns
    -------
    a : float64
        Coefficient of the equation of the plane.
    b : float64
        Coefficient of the equation of the plane.
    c : float64
        Coefficient of the equation of the plane.
    d : float64
        Coefficient of the equation of the plane.

    Examples
    --------
    >>> N1 = [2.298354000, 5.161785000, 7.971898000]
    >>> N2 = [1.885657000, 4.804777000, 6.183726000]
    >>> N3 = [1.747515000, 6.960963000, 7.932784000]

    >>> a, b, c, d = find_eq_of_plane(N1, N2, N3)

    >>> a
    -3.231203733528
    >>> b
    -0.9688526458499996
    >>> c
    0.9391692927779998
    >>> d
    -4.940497273569501

    """
    x = np.asarray(x, dtype=np.float64)
    y = np.asarray(y, dtype=np.float64)
    z = np.asarray(z, dtype=np.float64)

    xz = z - x
    xy = y - x

    cross_vector = np.cross(xz, xy)
    a, b, c = cross_vector

    d = np.dot(cross_vector, z)

    return a, b, c, d



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from matplotlib import pyplot as plt


class Plot:
    """
    Relationship plot between Zeta and Sigma parameters.

    Parameters
    ----------
    args[0] : list
        List of data set 1 (data1).
    args[0] = list
        List of data set 2 (data2).
    name1 = str, optional
        Name of data set 1.
    name2 = str, optional
        Name of data set 2.

    Examples
    --------
    >>> data1 = [1, 2, 3, 4, 5]
    >>> data2 = [1, 2, 3, 4, 5]

    >>> test = Plot(data1, data2, name1="Data 1", name2="Data 2")
    >>> test.add_point()
    >>> test.add_text()
    >>> test.add_legend()
    >>> test.show_plot()

    """
    def __init__(self, *args, name1="Var1", name2="Var2"):
        try:
            self.data1 = args[0]
        except NameError:
            raise NameError("data1 is not specified")

        try:
            self.data2 = args[1]
        except NameError:
            raise NameError("data2 is not specified")

        self.name1 = name1
        self.name2 = name2

        self.start_plot()
        self.config_plot()
        self.set_label()

    def start_plot(self):
        """
        Start plot.

        """
        self.ax = plt.subplot()

    def add_point(self):
        """
        Add all atoms to show in figure.

        """
        for i in range(len(self.data1)):
            self.ax.scatter(self.data1, self.data2, label=f'Complex {i + 1}')

    def add_text(self):
        """
        Added text to show in figure.

        """
        for i in range(len(self.data1)):
            self.ax.text(self.data1[i] + 0.2, self.data2[i] + 0.2, i + 1, fontsize=9)

    def add_legend(self):
        """
        Add legend to show in figure.

        """
        # Put a legend below current axis
        self.ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.1),
                       fancybox=True, shadow=True, ncol=5)

    def config_plot(self):
        """
        Config structure of figure.

        """
        # Shrink current axis's height by 10% on the bottom
        box = self.ax.get_position()
        self.ax.set_position([box.x0, box.y0 + box.height * 0.1, box.width, box.height * 0.9])

    def set_label(self):
        """
        Set title of figure and axis labels.

        """
        plt.title(f"Relationship plot between {self.name1} and {self.name2}")
        plt.xlabel(f"{self.name1}")
        plt.ylabel(f"{self.name2}")

    @staticmethod
    def save_img(save="Image_saved_by_OctaDist", file="png"):
        """
        Save figure as an image.

        Parameters
        ----------
        save : str
            Name of image file.
            Default value is "Complex_saved_by_OctaDist".
        file : file
            Image type.
            Default value is "png".

        """
        plt.savefig(f"{save}.{file}")

    @staticmethod
    def show_plot():
        """
        Show plot.

        """
        plt.show()




# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

__author__ = "R. Ketkaew, Y. Tantirungrotechai, D. J. Harding, P. Harding, and M. Marchivie"
__author_full__ = "Rangsiman Ketkaew, Yuthana Tantirungrotechai, David J. Harding, " \
                  "Phimphaka Harding, and Mathieu Marchivie"
__maintainer__ = "Rangsiman Ketkaew"
__copyright__ = "OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al."
__license__ = "GNU v3"
__credit__ = "OctaDist Development Team"
__email__ = "rangsiman1993@gmail.com"
__version__ = "2.5.4.5"
__revision__ = "2019.2545"
__release__ = "June 2019"
__status__ = "stable"
__title__ = "Octahedral Distortion Calculator"
__description__ = "OctaDist: A tool for calculating distortion parameters in coordination complexes."
__doc__ = "https://octadist.github.io/manual.html"
__website__ = "https://octadist.github.io"
__github__ = "https://github.com/OctaDist/OctaDist"

__all__ = \
    ['Run',
     'calc',
     'draw',
     'elements',
     'linear',
     'molecule',
     'plane',
     'plot',
     'popup',
     'projection',
     'structure',
     'tools',
     'util',
     # -----------------------
     'CalcDistortion',
     'DrawComplex',
     'DrawProjection',
     'DrawTwistingPlane',
     'check_atom',
     'check_radii',
     'check_color',
     'angle_sign',
     'angle_btw_vectors',
     'angle_btw_planes',
     'triangle_area',
     'count_line',
     'extract_coord',
     'find_metal',
     'extract_octa',
     'is_xyz',
     'is_gaussian',
     'is_nwchem',
     'is_orca',
     'is_qchem',
     'get_coord_xyz',
     'get_coord_gaussian',
     'get_coord_nwchem',
     'get_coord_orca',
     'get_coord_qchem',
     'find_eq_of_plane',
     'Plot',
     'project_atom_onto_line',
     'project_atom_onto_plane',
     'DataComplex',
     'StructParam',
     'SurfaceArea',
     'CalcJahnTeller',
     'CalcRMSD',
     'find_bonds',
     'find_faces_octa',
     'example_atom',
     'example_coord',
     ]


# Bring sub-modules to top-level directory

from octadist import logo
from .logo import Icon_Base64

from octadist import src

from .src import __src__

from .src import calc
from .src import draw
from .src import elements
from .src import linear
from .src import molecule
from .src import plane
from .src import plot
from .src import popup
from .src import projection
from .src import structure
from .src import tools
from .src import util

# Bring function and methods to top-level directory

from .src.calc import CalcDistortion

from .src.draw import DrawComplex
from .src.draw import DrawProjection
from .src.draw import DrawTwistingPlane

from .src.elements import check_atom
from .src.elements import check_radii
from .src.elements import check_color

from .src.linear import angle_sign
from .src.linear import angle_btw_vectors
from .src.linear import angle_btw_planes
from .src.linear import triangle_area

from .src.molecule import count_line
from .src.molecule import extract_coord
from .src.molecule import find_metal
from .src.molecule import extract_octa
from .src.molecule import is_xyz
from .src.molecule import is_gaussian
from .src.molecule import is_nwchem
from .src.molecule import is_orca
from .src.molecule import is_qchem
from .src.molecule import get_coord_xyz
from .src.molecule import get_coord_gaussian
from .src.molecule import get_coord_nwchem
from .src.molecule import get_coord_orca
from .src.molecule import get_coord_qchem

from .src.plane import find_eq_of_plane

from .src.plot import Plot

from .src.projection import project_atom_onto_line
from .src.projection import project_atom_onto_plane

from .src.structure import DataComplex
from .src.structure import StructParam
from .src.structure import SurfaceArea

from .src.tools import CalcJahnTeller
from .src.tools import CalcRMSD

from .src.util import find_bonds
from .src.util import find_faces_octa

example_atom = ['Fe', 'O', 'O', 'N', 'N', 'N', 'N']

example_coord = [[2.298354000, 5.161785000, 7.971898000],
                 [1.885657000, 4.804777000, 6.183726000],
                 [1.747515000, 6.960963000, 7.932784000],
                 [4.094380000, 5.807257000, 7.588689000],
                 [0.539005000, 4.482809000, 8.460004000],
                 [2.812425000, 3.266553000, 8.131637000],
                 [2.886404000, 5.392925000, 9.848966000]]




# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np
from scipy.spatial import distance

from octadist.src import linear, plane, projection


class CalcDistortion:
    """
    Calculate metal-ligand bond distance and return value in Angstrom.

    Parameters
    ----------
    coord : array_like
        Atomic coordinates of octahedral structure.

    Examples
    --------
    >>> coord = [[2.298354000, 5.161785000, 7.971898000],  # <- Metal atom
                 [1.885657000, 4.804777000, 6.183726000],
                 [1.747515000, 6.960963000, 7.932784000],
                 [4.094380000, 5.807257000, 7.588689000],
                 [0.539005000, 4.482809000, 8.460004000],
                 [2.812425000, 3.266553000, 8.131637000],
                 [2.886404000, 5.392925000, 9.848966000]]

    >>> test = CalcDistortion(coord)
    >>> sigma = test.sigma

    >>> sigma
    47.926528379270124

    """

    def __init__(self, coord):
        self.coord = coord

        if type(self.coord) == np.ndarray:
            pass
        else:
            self.coord = np.asarray(self.coord, dtype=np.float64)

        self.bond_dist = []
        self.d_mean = 0
        self.diff_dist = []
        self.zeta = 0
        self.delta = 0
        self.cis_angle = 0
        self.trans_angle = 0
        self.sigma = 0
        self.allTheta = []
        self.theta = 0
        self.theta_min = 0
        self.theta_max = 0
        self.non_octa = False

        self.calc_d_bond()
        self.calc_d_mean()
        self.calc_bond_angle()
        self.calc_zeta()
        self.calc_delta()
        self.calc_sigma()
        self.calc_theta()
        self.calc_theta_min()
        self.calc_theta_max()

    def calc_d_bond(self):
        """
        Calculate metal-ligand bond distance and return value in Angstrom.

        Parameters
        ----------
        coord : array_like
            Atomic coordinates of octahedral structure.

        See Also
        --------
        calc_d_mean : Calculate mean metal-ligand bond length.

        """
        self.bond_dist = [distance.euclidean(self.coord[0], self.coord[i]) for i in range(1, 7)]
        self.bond_dist = np.asarray(self.bond_dist, dtype=np.float64)

    def calc_d_mean(self):
        """
        Calculate mean distance parameter and return value in Angstrom.

        Parameters
        ----------
        coord : array_like
            Atomic coordinates of octahedral structure.

        See Also
        --------
        calc_d_bond : Calculate metal-ligand bonds length.

        """
        self.d_mean = np.mean(self.bond_dist)

    def calc_bond_angle(self):
        """
        Calculate 12 cis and 3 trans unique angles in octahedral structure.

        Parameters
        ----------
        coord : array_like
            Atomic coordinates of octahedral structure.

        See Also
        --------
        calc_sigma : Calculate Sigma parameter.

        """
        all_angle = []
        for i in range(1, 7):
            for j in range(i + 1, 7):
                vec1 = self.coord[i] - self.coord[0]
                vec2 = self.coord[j] - self.coord[0]
                angle = linear.angle_btw_vectors(vec1, vec2)
                all_angle.append(angle)

        # Sort the angle from the lowest to the highest
        sorted_angle = sorted(all_angle)
        self.cis_angle = [sorted_angle[i] for i in range(12)]
        self.trans_angle = [sorted_angle[i] for i in range(12, 15)]

    def calc_zeta(self):
        """
        Calculate Zeta parameter and return value in Angstrom.

        Parameters
        ----------
        coord : array_like
            Atomic coordinates of octahedral structure.

        See Also
        --------
        calc_d_bond : Calculate metal-ligand bonds length.
        calc_d_mean : Calculate mean metal-ligand bond length.

        Notes
        -----
        \zeta = \sum_{i=1}^{6}\left | d_{i} - d_{mean}  \right |

        where :math:`d_{i}` is individual M-X bond distance and
        :math:`d_{mean}` is mean metal-ligand bond distance.

        References
        ----------
        Phys. Rev. B 85, 064114

        """
        diff_dist = [abs(self.bond_dist[i] - self.d_mean) for i in range(6)]
        self.diff_dist = np.asarray(diff_dist, dtype=np.float64)

        self.zeta = np.sum(self.diff_dist)

    def calc_delta(self):
        """
        Calculate Delta parameter, also known as Tilting distortion parameter.

        Parameters
        ----------
        coord : array_like
            Atomic coordinates of octahedral structure.

        See Also
        --------
        calc_d_bond : Calculate metal-ligand bonds length.
        calc_d_mean : Calculate mean metal-ligand bond length.

        Notes
        -----
        \Delta = \frac{1}{6} \sum_{i=1}^{6}(\frac{d_{i} - d_{mean}}{d_{mean}})^2

        where :math:`d_{i}` is individual M-X bond distance and
        :math:`d_{mean}` is mean metal-ligand bond distance.

        References
        ----------
        DOI: 10.1107/S0108768103026661
        Acta Cryst. (2004). B60, 10-20

        """
        delta = sum(pow((self.bond_dist[i] - self.d_mean) / self.d_mean, 2) for i in range(6))
        self.delta = delta / 6

    def calc_sigma(self):
        """
        Calculate Sigma parameter and return value in degree.

        Parameters
        ----------
        coord : array_like
            Atomic coordinates of octahedral structure.

        See Also
        --------
        calc_bond_angle : Calculate bond angles between ligand-metal-ligand.

        Notes
        -----
        \Sigma = \sum_{i=1}^{12}\left | 90 - \phi_{i}  \right |

        where :math:`\phi_{i}` in individual cis angle.

        References
        ----------
        J. K. McCusker et al. Inorg. Chem. 1996, 35, 2100.

        """
        self.sigma = sum(abs(90.0 - self.cis_angle[i]) for i in range(12))

    def calc_theta(self):
        """
        Calculate Theta parameter and value in degree.

        Parameters
        ----------
        coord : array_like
            Atomic coordinates of octahedral structure.

        See Also
        --------
        calc_theta_min :
            Calculate minimum Theta parameter.
        calc_theta_max :
            Calculate maximum Theta parameter.
        octadist.src.linear.angle_btw_vectors :
            Calculate cosine angle between two vectors.
        octadist.src.linear.angle_sign
            Calculate cosine angle between two vectors sensitive to CW/CCW direction.
        octadist.src.plane.find_eq_of_plane :
            Find the equation of the plane.
        octadist.src.projection.project_atom_onto_plane :
            Orthogonal projection of point onto the plane.

        Notes
        -----
        \Theta = \sum_{i=1}^{24}\left | 60 - \theta_{i}  \right |

        where :math:`\theta_{i}` is individual angle between two vectors of two twisting face.

        References
        ----------
        M. Marchivie et al.
        Acta Crystal-logr. Sect. B Struct. Sci. 2005, 61, 25.

        """
        ligands = list(self.coord[1:])

        TM = self.coord[0]
        N1 = self.coord[1]
        N2 = self.coord[2]
        N3 = self.coord[3]
        N4 = self.coord[4]
        N5 = self.coord[5]
        N6 = self.coord[6]

        # Vector from metal to ligand atom
        TMN1 = N1 - TM
        TMN2 = N2 - TM
        TMN3 = N3 - TM
        TMN4 = N4 - TM
        TMN5 = N5 - TM
        TMN6 = N6 - TM

        ligands_vec = [TMN1, TMN2, TMN3, TMN4, TMN5, TMN6]

        ###########################################
        # Determine the order of atoms in complex #
        ###########################################

        max_angle = self.trans_angle[0]

        # This loop is used to identify which N is in line with N1
        def_change = 6
        for n in range(6):
            test = linear.angle_btw_vectors(ligands_vec[0], ligands_vec[n])
            if test > (max_angle - 1):
                def_change = n

        test_max = 0
        new_change = 0
        for n in range(6):
            test = linear.angle_btw_vectors(ligands_vec[0], ligands_vec[n])
            if test > test_max:
                test_max = test
                new_change = n

        # self.non_octa is used to identify the type of octahedral structure
        if def_change != new_change:
            self.non_octa = True
            def_change = new_change

        tp = ligands[4]
        ligands[4] = ligands[def_change]
        ligands[def_change] = tp

        N1 = ligands[0]
        N2 = ligands[1]
        N3 = ligands[2]
        N4 = ligands[3]
        N5 = ligands[4]
        N6 = ligands[5]

        TMN1 = N1 - TM
        TMN2 = N2 - TM
        TMN3 = N3 - TM
        TMN4 = N4 - TM
        TMN5 = N5 - TM
        TMN6 = N6 - TM

        ligands_vec = [TMN1, TMN2, TMN3, TMN4, TMN5, TMN6]

        # This loop is used to identify which N is in line with N2
        for n in range(6):
            test = linear.angle_btw_vectors(ligands_vec[1], ligands_vec[n])
            if test > (max_angle - 1):
                def_change = n

        # This loop is used to identify which N is in line with N2
        test_max = 0
        for n in range(6):
            test = linear.angle_btw_vectors(ligands_vec[1], ligands_vec[n])
            if test > test_max:
                test_max = test
                new_change = n

        if def_change != new_change:
            self.non_octa = True
            def_change = new_change

        # Swapping the atom (n+1) just identified above with N6
        tp = ligands[5]
        ligands[5] = ligands[def_change]
        ligands[def_change] = tp

        # New atom order is stored into the N1 - N6 lists
        N1 = ligands[0]
        N2 = ligands[1]
        N3 = ligands[2]
        N4 = ligands[3]
        N5 = ligands[4]
        N6 = ligands[5]

        TMN1 = N1 - TM
        TMN2 = N2 - TM
        TMN3 = N3 - TM
        TMN4 = N4 - TM
        TMN5 = N5 - TM
        TMN6 = N6 - TM

        ligands_vec = [TMN1, TMN2, TMN3, TMN4, TMN5, TMN6]

        # This loop is used to identify which N is in line with N3
        for n in range(6):
            test = linear.angle_btw_vectors(ligands_vec[2], ligands_vec[n])
            if test > (max_angle - 1):
                def_change = n

        # This loop is used to identify which N is in line with N3
        test_max = 0
        for n in range(6):
            test = linear.angle_btw_vectors(ligands_vec[2], ligands_vec[n])
            if test > test_max:
                test_max = test
                new_change = n

        if def_change != new_change:
            self.non_octa = True
            def_change = new_change

        # Swapping of the atom (n+1) just identified above with N4
        tp = ligands[3]
        ligands[3] = ligands[def_change]
        ligands[def_change] = tp

        # New atom order is stored into the N1 - N6 lists
        N1 = ligands[0]
        N2 = ligands[1]
        N3 = ligands[2]
        N4 = ligands[3]
        N5 = ligands[4]
        N6 = ligands[5]

        #####################################################
        # Calculate the Theta parameter ans its derivatives #
        #####################################################

        eqOfPlane = []
        indiTheta = []

        # loop over 8 faces
        for proj in range(8):
            a, b, c, d = plane.find_eq_of_plane(N1, N2, N3)
            eqOfPlane.append([a, b, c, d])

            # Project M, N4, N5, and N6 onto the plane defined by N1, N2, and N3
            TMP = projection.project_atom_onto_plane(TM, a, b, c, d)
            N4P = projection.project_atom_onto_plane(N4, a, b, c, d)
            N5P = projection.project_atom_onto_plane(N5, a, b, c, d)
            N6P = projection.project_atom_onto_plane(N6, a, b, c, d)

            VTh1 = N1 - TMP
            VTh2 = N2 - TMP
            VTh3 = N3 - TMP
            VTh4 = N4P - TMP
            VTh5 = N5P - TMP
            VTh6 = N6P - TMP

            a12 = linear.angle_btw_vectors(VTh1, VTh2)
            a13 = linear.angle_btw_vectors(VTh1, VTh3)
            if a12 < a13:
                direction = np.cross(VTh1, VTh2)
            else:
                direction = np.cross(VTh3, VTh1)

            theta1 = linear.angle_sign(VTh1, VTh4, direction)
            theta2 = linear.angle_sign(VTh4, VTh2, direction)
            theta3 = linear.angle_sign(VTh2, VTh5, direction)
            theta4 = linear.angle_sign(VTh5, VTh3, direction)
            theta5 = linear.angle_sign(VTh3, VTh6, direction)
            theta6 = linear.angle_sign(VTh6, VTh1, direction)

            indiTheta.append([theta1, theta2, theta3, theta4, theta5, theta6])

            sum_theta = sum(abs(indiTheta[proj][i] - 60) for i in range(6))

            self.allTheta.append(sum_theta)

            tp = N2
            N2 = N4
            N4 = N6
            N6 = N3
            N3 = tp

            # If the proj = 3, permutation face will be switched from N1N2N3
            # to N1N4N2, to N1N6N4, then to N1N3N6, and then back to N1N2N3
            if proj == 3:
                tp = N1
                N1 = N5
                N5 = tp
                tp = N2
                N2 = N6
                N6 = tp
                tp = N3
                N3 = N4
                N4 = tp

            # End of the loop that calculate the 8 projections.

        self.theta = sum(self.allTheta[i] for i in range(8)) / 2

    def calc_theta_min(self):
        """
        Calculate minimum Theta parameter and return value in degree.

        Parameters
        ----------
        allTheta : list
            List of individual Theta angles.

        """
        sorted_theta = sorted(self.allTheta)
        self.theta_min = sum(sorted_theta[i] for i in range(4))

    def calc_theta_max(self):
        """
        Calculate maximum Theta parameter and return value in degree.

        Parameters
        ----------
        allTheta : list
            List of individual Theta angles.

        """
        sorted_theta = sorted(self.allTheta)
        self.theta_max = sum(sorted_theta[i] for i in range(4, 8))



#!/usr/bin/env python

# Metadata file for distribution of OctaDist software on PyPI
# https://packaging.python.org/guides/distributing-packages-using-setuptools/

import setuptools

__version__ = "2.5.4.5"

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

description = "OctaDist: A tool for calculating distortion parameters in coordination complexes."

setuptools.setup(
    name="octadist",
    version=__version__,
    author="Rangsiman Ketkaew",
    author_email="rangsiman1993@gmail.com",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://octadist.github.io",
    download_url="https://github.com/OctaDist/OctaDist/releases",
    project_urls={
        'Documentation': 'https://octadist.readthedocs.io/en/latest/',
        'Source': 'https://github.com/OctaDist/OctaDist',
        'Tracker': 'https://github.com/OctaDist/OctaDist/issues',
    },
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'rmsd'
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Other Audience",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Multimedia :: Graphics :: 3D Modeling",
    ],
    keywords=[
        'chemistry'
        'crystallography',
        'inorganic chemistry',
        'coordination complex'
        'octahedral distortion',
    ],
    python_requires='>=3',
)



# OctaDist  Copyright (C) 2019  Rangsiman Ketkaew et al.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from operator import itemgetter

import numpy as np
from scipy.spatial import distance

from octadist.src import elements, popup


def count_line(file=None):
    """
    Count lines in an input file.

    Parameters
    ----------
    file : str
        Absolute or full path of input file.

    Returns
    -------
    i + 1 : int
        Number of line in file.

    Examples
    --------
    >>> file = "/home/Jack/[Fe(1-bpp)2][BF4]2-HS.xyz"
    >>> line = count_line(file)

    >>> line
    27

    """
    if file is None:
        raise TypeError("count_line needs one argument: input file")

    with open(file) as f:
        for i, l in enumerate(f):
            pass

    return i + 1


def extract_coord(file=None):
    """
    Check file type, read data, extract atom and coord from an input file.

    Parameters
    ----------
    file : str
        User input filename.

    Returns
    -------
    atom : list
        Full atomic labels of complex.
    coord : ndarray
        Full atomic coordinates of complex.

    See Also
    --------
    octadist.main.OctaDist.open_file :
        Open file dialog and to browse input file.
    octadist.main.OctaDist.search_coord :
        Search octahedral structure in complex.

    Notes
    -----
    The following is the file type that OctaDist supports:

    - ``XYZ``
    - ``Gaussian``
    - ``NWChem``
    - ``ORCA``
    - ``Q-Chem``

    Examples
    --------
    >>> file = "/home/Jack/[Fe(1-bpp)2][BF4]2-HS.xyz"
    >>> atom, coord = extract_coord(file)

    >>> atom
    ['Fe', 'N', 'N', 'N', 'N', 'N', 'N']

    >>> coord
    [[-1.95348286e+00,  4.51770478e+00,  1.47855811e+01],
     [-1.87618286e+00,  4.48070478e+00,  1.26484811e+01],
     [-3.90128286e+00,  5.27750478e+00,  1.40814811e+01],
     [-4.88286000e-03,  3.69060478e+00,  1.42392811e+01],
     [-2.18698286e+00,  4.34540478e+00,  1.69060811e+01],
     [-1.17538286e+00,  6.38340478e+00,  1.56457811e+01],
     [-2.75078286e+00,  2.50260478e+00,  1.51806811e+01]]

    """
    if file is None:
        raise TypeError("extract_coord needs one argument: input file")

    atom = []
    coord = np.array([])
    check = True

    # Check file extension
    if file.endswith(".xyz"):
        if is_xyz(file):
            atom, coord = get_coord_xyz(file)

        else:
            ftype = "XYZ"
            popup.err_invalid_ftype(ftype)
            check = False

    elif file.endswith(".out") or file.endswith(".log"):
        if is_gaussian(file):
            atom, coord = get_coord_gaussian(file)
        elif is_nwchem(file):
            atom, coord = get_coord_nwchem(file)
        elif is_orca(file):
            atom, coord = get_coord_orca(file)
        elif is_qchem(file):
            atom, coord = get_coord_qchem(file)
        else:
            check = False
    else:
        popup.err_wrong_format()
        check = False

    if check:
        # Remove empty string in list
        atom = list(filter(None, atom))
        return atom, coord
    else:
        return atom, coord


def find_metal(atom=None, coord=None):
    """
    Count the number of metal center atom in complex.

    Parameters
    ----------
    atom : list
        Full atomic labels of complex.
    coord : array_like
        Full atomic coordinates of complex.

    Returns
    -------
    total_metal : int
        The total number of metal center atom.
    atom_metal : list
        Atomic labels of metal center atom.
    coord_metal : ndarray
        Atomic coordinates of metal center atom.

    See Also
    --------
    octadist.src.elements.check_atom :
        Convert atomic number to atomic symbol and vice versa.

    Examples
    --------
    >>> atom = ['Fe', 'N', 'N', 'N', 'N', 'N', 'N']

    >>> coord = [[-1.95348286e+00,  4.51770478e+00,  1.47855811e+01],
                 [-1.87618286e+00,  4.48070478e+00,  1.26484811e+01],
                 [-3.90128286e+00,  5.27750478e+00,  1.40814811e+01],
                 [-4.88286000e-03,  3.69060478e+00,  1.42392811e+01],
                 [-2.18698286e+00,  4.34540478e+00,  1.69060811e+01],
                 [-1.17538286e+00,  6.38340478e+00,  1.56457811e+01],
                 [-2.75078286e+00,  2.50260478e+00,  1.51806811e+01]]

    >>> total_metal, atom_metal, coord_metal = find_metal(atom, coord)

    >>> total_metal
    1

    >>> atom_metal
    ['Fe']

    >>> coord_metal
    [[-1.95348286  4.51770478 14.78558113]]

    """
    if atom is None or coord is None:
        raise TypeError("find_metal needs two arguments: atom, coord")

    total_metal = 0
    atom_metal = []
    coord_metal = []

    for i in range(len(atom)):
        number = elements.check_atom(atom[i])

        if 21 <= number <= 30 or \
                39 <= number <= 48 or \
                57 <= number <= 80 or \
                89 <= number <= 109:

            total_metal += 1
            atom_metal.append(atom[i])
            coord_metal.append(coord[i])

    coord_metal = np.asarray(coord_metal, dtype=np.float64)

    return total_metal, atom_metal, coord_metal


def extract_octa(atom=None, coord=None, metal=1, cutoff_metal_ligand=2.8):
    """
    Search the octahedral structure in complex and return atoms and coordinates.

    Parameters
    ----------
    atom : list
        Full atomic labels of complex.
    coord : array_like
        Full atomic coordinates of complex.
    metal : int
        Number of metal atom that will be taken as center atom for
        finding atomic coordinates of octahedral structure of interest.
        Default value is 1.
    cutoff_metal_ligand : float, optional
        Cutoff distance for screening metal-ligand bond.
        Default value is 2.8.

    Returns
    -------
    atom_octa : list
        Atomic labels of octahedral structure.
    coord_octa : ndarray
        Atomic coordinates of octahedral structure.

    See Also
    --------
    find_metal :
        Find metals in complex.
    octadist.main.OctaDist.search_coord :
        Search octahedral structure in complex.

    Examples
    --------
    >>> atom = ['Fe', 'N', 'N', 'N', 'N', 'N', 'N']

    >>> coord = [[-1.95348286e+00,  4.51770478e+00,  1.47855811e+01],
                 [-1.87618286e+00,  4.48070478e+00,  1.26484811e+01],
                 [-3.90128286e+00,  5.27750478e+00,  1.40814811e+01],
                 [-4.88286000e-03,  3.69060478e+00,  1.42392811e+01],
                 [-2.18698286e+00,  4.34540478e+00,  1.69060811e+01],
                 [-1.17538286e+00,  6.38340478e+00,  1.56457811e+01],
                 [-2.75078286e+00,  2.50260478e+00,  1.51806811e+01]]

    >>> atom_octa, coord_octa = extract_octa(atom, coord)

    >>> atom_octa
    ['Fe', 'N', 'N', 'N', 'N', 'N', 'N']

    >>> coord_octa
    [[-1.95348286e+00,  4.51770478e+00,  1.47855811e+01],
     [-1.87618286e+00,  4.48070478e+00,  1.26484811e+01],
     [-3.90128286e+00,  5.27750478e+00,  1.40814811e+01],
     [-4.88286000e-03,  3.69060478e+00,  1.42392811e+01],
     [-2.18698286e+00,  4.34540478e+00,  1.69060811e+01],
     [-1.17538286e+00,  6.38340478e+00,  1.56457811e+01],
     [-2.75078286e+00,  2.50260478e+00,  1.51806811e+01]]

    """
    if atom is None or coord is None:
        raise TypeError("find_metal needs two arguments: atom and coord")

    # Count the number of metal center atom
    total_metal, atom_metal, coord_metal = find_metal(atom, coord)

    if metal <= 0:
        raise ValueError("index of metal must be positive integer")

    elif metal > total_metal:
        raise ValueError("user-defined index of metal is greater than "
                         "the total metal in complex.")

    metal_index = metal - 1
    dist_list = []

    for i in range(len(list(atom))):
        dist = distance.euclidean(coord_metal[metal_index], coord[i])
        if dist <= cutoff_metal_ligand:
            dist_list.append([atom[i], coord[i], dist])

    # sort list of tuples by distance in ascending order
    dist_list.sort(key=itemgetter(2))

    # Keep only first 7 atoms
    dist_list = dist_list[:7]
    atom_octa, coord_octa, dist = zip(*dist_list)

    atom_octa = list(atom_octa)
    coord_octa = np.asarray(coord_octa, dtype=np.float64)

    return atom_octa, coord_octa


def is_xyz(f):
    """
    Check if the input file is .xyz file format.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    bool : bool
        If file is XYZ file, return True.

    See Also
    --------
    get_coord_xyz :
        Find atomic coordinates of molecule from XYZ file.

    Notes
    -----
    XYZ file format is like following:

    >>> 
    <number of atom>
    comment
    <index 0> <X> <Y> <Z>
    <index 1> <X> <Y> <Z>
    ...
    <index 6> <X> <Y> <Z>

    Examples
    --------
    >>> example.xyz 
    7
    Comment: From Excel file
    Fe  6.251705    9.063211    5.914842
    N   8.15961     9.066456    5.463087
    N   6.749414    10.457551   7.179682
    N   5.709997    10.492955   4.658257
    N   4.350474    9.106286    6.356091
    O   5.789096    7.796326    4.611355
    O   6.686381    7.763872    7.209699

    >>> is_xyz(example.xyz)
    True

    """
    file = open(f, "r")

    first_line = file.readline()

    # Check if the first line is integer
    try:
        int(first_line)
    except ValueError:
        return False

    if count_line(f) < 9:
        return False
    else:
        return True


def get_coord_xyz(f):
    """
    Get coordinate from .xyz file.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    atom : list
        Full atomic labels of complex.
    coord : ndarray
        Full atomic coordinates of complex.

    """
    file = open(f, "r")
    # read file from 3rd line
    line = file.readlines()[2:]
    file.close()

    atom = []
    for l in line:
        # read atom on 1st column and insert to list
        l_strip = l.strip()
        lst = l_strip.split(' ')[0]
        atom.append(lst)

    file = open(f, "r")
    coord = np.loadtxt(file, skiprows=2, usecols=[1, 2, 3])
    file.close()

    coord = np.asarray(coord, dtype=np.float64)

    return atom, coord


def is_gaussian(f):
    """
    Check if the input file is Gaussian file format.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    bool : bool
        If file is Gaussian output file, return True.

    See Also
    --------
    get_coord_gaussian :
        Find atomic coordinates of molecule from Gaussian file.

    Examples
    --------
    >>> gaussian.log
                                Standard orientation:
    ---------------------------------------------------------------------
    Center     Atomic      Atomic             Coordinates (Angstroms)
    Number     Number       Type             X           Y           Z
    ---------------------------------------------------------------------
         1         26           0        0.000163    1.364285   -0.000039
         2          8           0        0.684192    0.084335   -1.192008
         3          8           0       -0.683180    0.083251    1.191173
         4          7           0        1.639959    1.353157    1.006941
         5          7           0       -0.563377    2.891083    1.435925
    ...

    """
    gaussian_file = open(f, "r")
    nline = gaussian_file.readlines()

    for i in range(len(nline)):
        if "Standard orientation:" in nline[i]:
            return True

    return False


def get_coord_gaussian(f):
    """
    Extract XYZ coordinate from Gaussian output file.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    atom : list
        Full atomic labels of complex.
    coord : ndarray
        Full atomic coordinates of complex.

    """
    gaussian_file = open(f, "r")
    nline = gaussian_file.readlines()

    start = 0
    end = 0
    atom, coord = [], []
    for i in range(len(nline)):
        if "Standard orientation:" in nline[i]:
            start = i

    for i in range(start + 5, len(nline)):
        if "---" in nline[i]:
            end = i
            break

    for line in nline[start + 5: end]:
        data = line.split()
        data1 = int(data[1])
        coord_x = float(data[3])
        coord_y = float(data[4])
        coord_z = float(data[5])
        data1 = elements.check_atom(data1)
        atom.append(data1)
        coord.append([coord_x, coord_y, coord_z])

    gaussian_file.close()

    coord = np.asarray(coord, dtype=np.float64)

    return atom, coord


def is_nwchem(f):
    """
    Check if the input file is NWChem file format.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    bool : bool
        If file is NWChem output file, return True.

    See Also
    --------
    get_coord_nwchem :
        Find atomic coordinates of molecule from NWChem file.

    Examples
    --------
    >>> nwchem.out
      ----------------------
      Optimization converged
      ----------------------
    ...
    ...
     No.       Tag          Charge          X              Y              Z
    ---- ---------------- ---------- -------------- -------------- --------------
       1 Ru(Fragment=1)      44.0000    -3.04059115    -0.08558108    -0.07699482
       2 C(Fragment=1)        6.0000    -1.62704660     2.40971357     0.63980357
       3 C(Fragment=1)        6.0000    -0.61467778     0.59634595     1.68841986
       4 C(Fragment=1)        6.0000     0.31519183     1.41684566     2.30745116
       5 C(Fragment=1)        6.0000     0.28773462     2.80126911     2.08006241
    ...

    """
    nwchem_file = open(f, "r")
    nline = nwchem_file.readlines()

    for i in range(len(nline)):
        if "No. of atoms" in nline[i]:
            if not int(nline[i].split()[4]):
                return False

    for j in range(len(nline)):
        if "Optimization converged" in nline[j]:
            return True

    return False


def get_coord_nwchem(f):
    """
    Extract XYZ coordinate from NWChem output file.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    atom : list
        Full atomic labels of complex.
    coord : ndarray
        Full atomic coordinates of complex.

    """
    nwchem_file = open(f, "r")
    nline = nwchem_file.readlines()

    start = 0
    end = 0
    atom, coord = [], []
    for i in range(len(nline)):
        if "Optimization converged" in nline[i]:
            start = i

    for i in range(len(nline)):
        if "No. of atoms" in nline[i]:
            end = int(nline[i].split()[4])

    start = start + 18
    end = start + end
    # The 1st line of coordinate is at 18 lines next to 'Optimization converged'
    for line in nline[start:end]:
        dat = line.split()
        dat1 = int(float(dat[2]))
        coord_x = float(dat[3])
        coord_y = float(dat[4])
        coord_z = float(dat[5])
        dat1 = elements.check_atom(dat1)
        atom.append(dat1)
        coord.append([coord_x, coord_y, coord_z])

    nwchem_file.close()

    coord = np.asarray(coord, dtype=np.float64)

    return atom, coord


def is_orca(f):
    """
    Check if the input file is ORCA file format.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    bool : bool
        If file is ORCA output file, return True.

    See Also
    --------
    get_coord_orca :
        Find atomic coordinates of molecule from ORCA file.

    Examples
    --------
    >>> orca.out
    ---------------------------------
    CARTESIAN COORDINATES (ANGSTROEM)
    ---------------------------------
      C      0.009657    0.000000    0.005576
      C      0.009657   -0.000000    1.394424
      C      1.212436   -0.000000    2.088849
      C      2.415214    0.000000    1.394425
      C      2.415214   -0.000000    0.005575
    ...

    """
    orca_file = open(f, "r")
    nline = orca_file.readlines()
    for i in range(len(nline)):
        if "CARTESIAN COORDINATES (ANGSTROEM)" in nline[i]:
            return True

    return False


def get_coord_orca(f):
    """
    Extract XYZ coordinate from ORCA output file.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    atom : list
        Full atomic labels of complex.
    coord : ndarray
        Full atomic coordinates of complex.

    """
    orca_file = open(f, "r")
    nline = orca_file.readlines()

    start = 0
    end = 0
    atom, coord = [], []
    for i in range(len(nline)):
        if "CARTESIAN COORDINATES (ANGSTROEM)" in nline[i]:
            start = i

    for i in range(start + 2, len(nline)):
        if "---" in nline[i]:
            end = i - 1
            break

    for line in nline[start + 2:end]:
        dat = line.split()
        dat1 = dat[0]
        coord_x = float(dat[1])
        coord_y = float(dat[2])
        coord_z = float(dat[3])
        atom.append(dat1)
        coord.append([coord_x, coord_y, coord_z])

    orca_file.close()

    coord = np.asarray(coord, dtype=np.float64)

    return atom, coord


def is_qchem(f):
    """
    Check if the input file is Q-Chem file format.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    bool : bool
        If file is Q-Chem output file, return True.

    See Also
    --------
    get_coord_qchem :
        Find atomic coordinates of molecule from Q-Chem file.

    Examples
    --------
    >>> qchem.out
    ******************************
    **  OPTIMIZATION CONVERGED  **
    ******************************
                               Coordinates (Angstroms)
        ATOM                X               Y               Z
         1  C         0.2681746845   -0.8206222796   -0.3704019386
         2  C        -1.1809302341   -0.5901746612   -0.6772716414
         3  H        -1.6636318262   -1.5373167851   -0.9496501352
         4  H        -1.2829834971    0.0829227646   -1.5389938241
         5  C        -1.9678565203    0.0191922768    0.5346693165
    ...

    """
    qchem_file = open(f, "r")
    nline = qchem_file.readlines()

    for i in range(len(nline)):
        if "OPTIMIZATION CONVERGED" in nline[i]:
            return True

    return False


def get_coord_qchem(f):
    """
    Extract XYZ coordinate from Q-Chem output file.

    Parameters
    ----------
    f : str
        User input filename.

    Returns
    -------
    atom : list
        Full atomic labels of complex.
    coord : ndarray
        Full atomic coordinates of complex.

    """
    orca_file = open(f, "r")
    nline = orca_file.readlines()

    start = 0
    end = 0
    atom, coord = [], []
    for i in range(len(nline)):
        if "OPTIMIZATION CONVERGED" in nline[i]:
            start = i

    for i in range(start + 5, len(nline)):
        if "Z-matrix Print:" in nline[i]:
            end = i - 1
            break

    for line in nline[start + 5:end]:
        dat = line.split()
        dat1 = dat[1]
        coord_x = float(dat[2])
        coord_y = float(dat[3])
        coord_z = float(dat[4])
        atom.append(dat1)
        coord.append([coord_x, coord_y, coord_z])

    orca_file.close()

    coord = np.asarray(coord, dtype=np.float64)

    return atom, coord
